import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions.normal import Normal
from torch.distributions.lowrank_multivariate_normal import LowRankMultivariateNormal
import torch.nn.functional as f
import numpy as np
import math
# from torch.distributions.multivariate_normal import MultivariateNormal
from torch.distributions.mixture_same_family import MixtureSameFamily
from torch.distributions.categorical import Categorical
from torch.distributions.independent import Independent
from torch.distributions.normal import Normal
import matplotlib.pyplot as plt
import cv2

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class MLP(nn.Module):

    def __init__(self, nin, nout, nh):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(nin, nh, bias=False),
            nn.LeakyReLU(0.2),
            # nn.ReLU(),
            nn.Linear(nh, nh),
            nn.LeakyReLU(0.2),
            # nn.ReLU(),
            nn.Linear(nh, nh),
            nn.LeakyReLU(0.2),
            # nn.ReLU(),
            nn.Linear(nh, nout, bias=False)
        )

    def forward(self, x):
        return self.net(x)

class AffineHalfFlow(nn.Module):
    def __init__(self, dim, parity, net_class=MLP, nh=20, scale=True, shift=True):
        super().__init__()
        self.dim = dim
        self.parity = parity
        if scale:
            self.s_cond = net_class(self.dim // 2, self.dim // 2, nh)
        if shift:
            self.t_cond = net_class(self.dim // 2, self.dim // 2, nh)
        self.W = nn.Linear(dim, dim, bias=False)


    def forward(self, x):
        x0, x1 = x[:, :, :(self.dim//2)], x[:, :, (self.dim//2):]
        if self.parity:
            x0, x1 = x1, x0
        s = self.s_cond(x0)
        t = self.t_cond(x0)
        z0 = x0
        z1 = torch.exp(s) * x1 + t
        if self.parity:
            z0, z1 = z1, z0
        z = torch.cat([z0, z1], dim=2)
        log_det = torch.sum(s, dim=[1,2])
        return z, log_det


    def backward(self, z):
        z0, z1 = z[:, :, :(self.dim // 2)], z[:, :, (self.dim // 2):]
        if self.parity:
            z0, z1 = z1, z0
        s = self.s_cond(z0)
        t = self.t_cond(z0)
        x0 = z0
        x1 = (z1 - t) * torch.exp(-s)
        if self.parity:
            x0, x1 = x1, x0
        x = torch.cat([x0, x1], dim=2)
        log_det = torch.sum(-s, dim=[1,2])
        return x, log_det

class NormalizingFlow(nn.Module):

    def __init__(self, hp, flows):
        super().__init__()
        self.hp = hp
        self.flows = nn.ModuleList(flows)

    def forward(self, x):
        m, _, _ = x.shape
        log_det = torch.zeros(m).to(device)
        zs = [x]
        for index, flow in enumerate(self.flows):
            x, ld = flow.forward(x)
            log_det += ld
            zs.append(x)
        return zs, log_det

    def backward(self, z):
        m, _, _ = z.shape
        log_det = torch.zeros(m).to(device)
        xs = [z]
        for flow in self.flows[::-1]:
            z, ld = flow.backward(z)
            log_det += ld
            xs.append(z)
        return xs, log_det


class NormalizingFlowModel(nn.Module):
    def __init__(self, hp, prior, flows,):
        super().__init__()
        self.prior = prior
        self.flow = NormalizingFlow(hp, flows)
        self.hp = hp

    def forward(self, x):
        zs, log_det = self.flow.forward(x[:,1:,:-10])
        if (self.hp.cov == "simple"):
            mu, D, logpi = self.prior(x[:, :-1, :])
            prior_logprob = gmm_sequence(zs[-1], mu, D, logpi)
        else:
            mu, D, logpi, U = self.prior(x[:, :-1, :])
            prior_logprob = gmm_Uout(zs[-1], mu, D, logpi, U, self.hp)
        return zs, prior_logprob, log_det

    def backward(self, z):
        xs, log_det = self.flow.backward(z)
        return xs, log_det

    def sample(self, x0, fig_path):

        if ("mnist" in fig_path):

            sample = torch.zeros((1, 28, 28)).to(device)
            sample[0, 0, :] = x0[0, 0, :28]

            for i in range(1, 28):
                data_with_taregt = torch.cat((sample[0, :i, :], x0[0, :i, 28:]), 1).unsqueeze(0)
                mu, sigma, logpi = self.forward(data_with_taregt)
                mix = Categorical(torch.exp(logpi))
                comp = Independent(Normal(mu, sigma), 1)
                gmm = MixtureSameFamily(mix, comp)
                row = gmm.sample()
                sample[0, i, :] = row[0, -1, :]

            im = sample.view((28, 28))
            # m = torch.nn.Sigmoid()
            # im = m(im)
            plt.imshow(im.cpu().detach().numpy(), cmap="gray")
            
        else:

            sample = torch.zeros((1, 32, 32 * 3)).to(device)
            sample[0, 0, :] = x0[0, 0, :32 * 3]

            for i in range(1, 32):
                data_with_taregt = torch.cat((sample[0, :i, :], x0[0, :i, 32 * 3:]), 1).unsqueeze(0)
                mu, sigma, logpi = self.prior(data_with_taregt)
                mix = Categorical(torch.exp(logpi))
                comp = Independent(Normal(mu, sigma), 1)
                gmm = MixtureSameFamily(mix, comp)
                row = gmm.sample()
                backwarded_row, _ = self.flow.backward(row)
                row = backwarded_row[-1]
                sample[0, i, :] = row[0, -1, :]

            im = sample.view((32, 32 * 3)).cpu().detach().numpy()
            im = im.astype(np.float32)
            list_im = []
            for i in range(32):
                for j in range(96):
                    list_im.append(im[i, j])
            im = np.array(list_im)

            count = 0
            image = np.zeros((32, 32, 3))
            for c in range(3):
                for h in range(32):
                    for w in range(32):
                        image[h, w, c] = im[count]
                        count += 1

            im = image
            im = (im - np.min(im)) / (np.max(im) - np.min(im))
            im = im.astype(np.float32)
            pixels = np.array(im)
            # image = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
            # pixels = np.array(image)
            plt.imshow(pixels)

        plt.savefig(fig_path)
        print("saved image at ", fig_path)


class FRMDN(object):
    def __init__(self, hp):

        if (hp.data_name == "mnist"):
            self.y_dim = 28
        else:
            self.y_dim = 32 * 3

        self.flows = [AffineHalfFlow(dim=self.y_dim, parity=i % 2) for i in range(hp.n_block)]
        # self.flows = [IdentityFlow(dim=self.y_dim, parity=i % 2) for i in range(hp.n_block)]

        if (hp.cov== "simple"):
            self.prior = RMDN(hp)
        elif (hp.cov == "output"):
            self.prior = RMDN_Uout(hp)
        else:
            self.prior = RMDN_Udirect(hp)
        self.flow_based = NormalizingFlowModel(hp, self.prior, self.flows)

class RMDN(nn.Module):
    def __init__(self, hp):
        super().__init__()
        if (hp.data_name == "mnist"):
            self.x_dim =  hp.class_number + 28
            self.y_dim = 28
            self.sequence_length = 28
        else:
            self.x_dim = hp.class_number + 32 * 3
            self.y_dim = 32 * 3
            self.sequence_length = 32
        self.gmm_linear = nn.Linear(hp.hidden_number, (2 * self.y_dim + 1) * hp.K, bias=True)
        self.rnn = nn.LSTM(input_size=self.x_dim, hidden_size=hp.hidden_number, num_layers=hp.hidden_layer, batch_first=True, bias=True)
        # torch.nn.init.xavier_uniform_(self.gmm_linear.weight)
        self.hp = hp

    def forward(self, inputs):

        self.sequence_length = inputs.shape[1]

        outs, _ = self.rnn(inputs)
        gmm_outs = self.gmm_linear(outs)

        stride = self.hp.K * self.y_dim

        mu = gmm_outs[:, :, :stride]
        mu = mu.view(-1, self.sequence_length, self.hp.K, self.y_dim)

        D = gmm_outs[:, :, stride: 2 * stride]
        D = D.view(-1, self.sequence_length, self.hp.K, self.y_dim)
        D = torch.exp(D) + 0.001
        # D = f.softplus(D) + 0.0001
        # D = torch.clamp(torch.sqrt(torch.exp(D)), min=0.1, max=1.9)

        pi = gmm_outs[:, :, 2 * stride:2 * stride +self.hp.K]
        pi = pi.view(-1, self.sequence_length, self.hp.K)
        logpi = F.log_softmax(pi, dim=-1)

        return mu, D, logpi

    def sample(self, x0, fig_path):

        if ("mnist" in fig_path):

            sample = torch.zeros((1, 28, 28)).to(device)
            sample[0, 0, :] = x0[0, 0, :28]

            for i in range(1, 28):
                data_with_taregt = torch.cat((sample[0, :i, :], x0[0, :i, 28:]), 1).unsqueeze(0)
                mu, sigma, logpi = self.forward(data_with_taregt)
                mix = Categorical(torch.exp(logpi))
                comp = Independent(Normal(mu, sigma), 1)
                gmm = MixtureSameFamily(mix, comp)
                row = gmm.sample()
                sample[0, i, :] = row[0, -1, :]

            im = sample.view((28, 28))
            plt.imshow(im.cpu().detach().numpy(), cmap="gray")

        else:

            sample = torch.zeros((1, 32, 32*3)).to(device)
            sample[0, 0, :] = x0[0, 0, :32*3]

            for i in range(1, 32):
                data_with_taregt = torch.cat((sample[0, :i, :], x0[0, :i, 32*3:]), 1).unsqueeze(0)
                mu, sigma, logpi = self.forward(data_with_taregt)
                mix = Categorical(torch.exp(logpi))
                comp = Independent(Normal(mu, sigma), 1)
                gmm = MixtureSameFamily(mix, comp)
                row = gmm.sample()
                sample[0, i, :] = row[0, -1, :]

            im = sample.view((32, 32*3)).cpu().detach().numpy()
            im = im.astype(np.float32)
            list_im = []
            for i in range(32):
                for j in range(96):
                    list_im.append(im[i,j])
            im = np.array(list_im)

            count = 0
            image = np.zeros((32, 32, 3))
            for c in range(3):
                for h in range(32):
                    for w in range(32):
                        image[h, w, c] = im[count]
                        count += 1

            im = image
            im = (im - np.min(im)) / (np.max(im) - np.min(im))
            im = im.astype(np.float32)
            image = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
            pixels = np.array(image)
            plt.imshow(pixels)

        plt.savefig(fig_path)
        print("saved image at ", fig_path)

def gmm_sequence(batch, mu, sigma, logpi, reduce=True):
    batch = batch[:, :, None, :]
    normal_dist = Normal(mu, sigma)
    g_log_probs = normal_dist.log_prob(batch)
    g_log_probs = logpi + torch.sum(g_log_probs, dim=-1)
    log_prob = torch.logsumexp(g_log_probs, -1)
    log_prob = torch.sum(log_prob, -1)
    if (reduce == True):
        return - torch.mean(log_prob)
    else:
        return log_prob

class RMDN_Uout(nn.Module):

    def __init__(self, hp):
        super().__init__()
        if (hp.data_name == "mnist"):
            self.x_dim =  hp.class_number + 28
            self.y_dim = 28
            self.sequence_length = 28
        else:
            self.x_dim = hp.class_number + 32 * 3
            self.y_dim = 32 * 3
            self.sequence_length = 32
        self.gmm_linear = nn.Linear(hp.hidden_number,(1 + ((2 + hp.dprim) * self.y_dim)) * hp.K, bias=True)
        self.rnn = nn.LSTM(input_size=self.x_dim, hidden_size=hp.hidden_number, num_layers=hp.hidden_layer, batch_first=True, bias=True)
        # torch.nn.init.xavier_uniform_(self.gmm_linear.weight)
        self.hp = hp

    def forward(self, inputs):

        self.sequence_length = inputs.shape[1]

        outs, _ = self.rnn(inputs)
        gmm_outs = self.gmm_linear(outs)

        stride = self.hp.K * self.y_dim

        mu = gmm_outs[:, :, :stride]
        mu = mu.view(-1, self.sequence_length, self.hp.K, self.y_dim)

        D = gmm_outs[:, :, stride: 2 * stride]
        D = D.view(-1, self.sequence_length, self.hp.K, self.y_dim)
        # D = torch.sqrt(torch.exp(D)) + 0.01
        # D = torch.exp(D) + 0.001
        D = f.softplus(D) + 0.0001
        # D = torch.clamp(torch.sqrt(torch.exp(D)), min=0.1, max=1.9)

        pi = gmm_outs[:, :, 2 * stride:2 * stride +self.hp.K]
        pi = pi.view(-1, self.sequence_length, self.hp.K)
        logpi = f.log_softmax(pi, dim=-1)

        U = gmm_outs[:, :, 2 * stride +self.hp.K:]
        U = U.view(-1, self.sequence_length, self.hp.K, self.y_dim, self.hp.dprim)

        return mu, D, logpi, U

class RMDN_Udirect(nn.Module):
    def __init__(self, hp):
        super().__init__()
        if (hp.data_name == "mnist"):
            self.x_dim =  hp.class_number + 28
            self.y_dim = 28
            self.sequence_length = 28
        else:
            self.x_dim = hp.class_number + 32 * 3
            self.y_dim = 32 * 3
            self.sequence_length = 32
        self.gmm_linear = nn.Linear(hp.hidden_number, (2 * self.y_dim + 1) * hp.K, bias=True)
        self.rnn = nn.LSTM(input_size=self.x_dim, hidden_size=hp.hidden_number, num_layers=hp.hidden_layer, batch_first=True, bias=True)
        # torch.nn.init.xavier_uniform_(self.gmm_linear.weight)
        self.hp = hp

        u_coefficient = 0.4
        self.U = torch.nn.Parameter(
            torch.from_numpy(u_coefficient * np.random.rand(hp.K, self.y_dim, hp.dprim)).float(),
            requires_grad=True)


    def forward(self, inputs):

        self.sequence_length = inputs.shape[1]

        outs, _ = self.rnn(inputs)
        gmm_outs = self.gmm_linear(outs)

        stride = self.hp.K * self.y_dim

        mu = gmm_outs[:, :, :stride]
        mu = mu.view(-1, self.sequence_length, self.hp.K, self.y_dim)

        D = gmm_outs[:, :, stride: 2 * stride]
        # D = D.view(1, -1, self.hp.K, self.y_dim)
        # D = torch.sqrt(torch.exp(D)) + 0.01
        # D = torch.clamp(torch.sqrt(torch.exp(D)), min=0.1, max=1.9)
        D = D.view(-1, self.sequence_length, self.hp.K, self.y_dim)
        # D = torch.sqrt(torch.exp(D)) + 0.01
        D = torch.exp(D) + 0.001
        # D = f.softplus(D) + 0.0001
        # D = torch.clamp(torch.sqrt(torch.exp(D)), min=0.1, max=1.9)

        pi = gmm_outs[:, :, 2 * stride:2 * stride +self.hp.K]
        pi = pi.view(-1, self.sequence_length, self.hp.K)
        logpi = f.log_softmax(pi, dim=-1)

        U = self.U.repeat(mu.shape[0], self.sequence_length, 1, 1, 1)#.unsqueeze(0)

        return mu, D, logpi, U


def gmm_Uout(batch, mu, D, logpi, U, hp, reduce=True):
    # # batch => [Batch, sequence length, y_dim]
    # # mu => [Batch, sequence length, gaussian, y_dim]
    # # D => [Batch, sequence length, gaussian, y_dim]
    # # logpi => [Batch, sequence length, gaussian]
    # # U => [Batch, sequence length, gaussian, y_dim, dprim]

    batch = batch[:, :, None, :]
    batch = batch.repeat(1, 1, hp.K, 1)

    b = U.shape[0]
    T = U.shape[1]
    K = U.shape[2]

    d = U.shape[-2]
    dprim = U.shape[-1]

    Y = batch - mu

    YD = torch.pow(Y * torch.sqrt(D), 2)
    YDY = torch.sum(YD, -1)

    Y = Y[:, :, :, None, :]

    YU = Y @ U
    YUUY = YU @ YU.transpose(-2, -1)

    exp_term = -0.5 * (YDY + YUUY.squeeze())
    const_term = -(d / 2) * np.log(2 * math.pi)

    Dinverse = torch.div(1, torch.sqrt(D))[:, :, :, None, :]
    Dinverse = Dinverse.repeat(1, 1, 1, dprim, 1)
    Utranspose = torch.transpose(U, -2, -1)
    UDinverse = Utranspose * Dinverse
    UDinverseTranspose = torch.transpose(UDinverse, -2, -1)
    UDU = torch.eye(dprim).to(device)[None, None, None, :, :].repeat(b, T, K, 1, 1) + (UDinverse @ UDinverseTranspose)
    det1 = torch.logdet(UDU)
    det2 = torch.sum(torch.log(D), -1)

    log_gmm = logpi + const_term + 0.5 * (det1 + det2) + exp_term

    log_sum_exp = torch.logsumexp(log_gmm, -1)
    sum_over_sequence_length = torch.sum(log_sum_exp, -1)

    if (reduce == True):
        return - torch.mean(sum_over_sequence_length)
    else:
        return sum_over_sequence_length