import numpy as np
import pickle

def logit(x):
    """
    Elementwise logit (inverse logistic sigmoid).
    :param x: numpy array
    :return: numpy array
    """
    return np.log(x / (1.0 - x))

def one_hot_encode(labels, n_labels):
    """
    Transforms numeric labels to 1-hot encoded labels. Assumes numeric labels are in the range 0, 1, ..., n_labels-1.
    """

    assert np.min(labels) >= 0 and np.max(labels) < n_labels

    y = np.zeros([labels.size, n_labels])
    y[range(labels.size), labels] = 1

    return y

class CIFAR10:
    """
    The CIFAR-10 dataset.
    https://www.cs.toronto.edu/~kriz/cifar.html
    """

    alpha = 0.05

    class Data:
        """
        Constructs the dataset.
        """

        def __init__(self, x, l, logit, flip, dequantize, rng):

            D = int(x.shape[1] / 3)                                 # number of pixels
            x = self._dequantize(x, rng) if dequantize else x  # dequantize
            x = self._logit_transform(x) if logit else x       # logit
            x = self._flip_augmentation(x) if flip else x      # flip
            self.x = x                                         # pixel values
            self.r = self.x[:, :D]                             # red component
            self.g = self.x[:, D:2*D]                          # green component
            self.b = self.x[:, 2*D:]                           # blue component
            self.labels = np.hstack([l, l]) if flip else l     # numeric labels
            self.y = one_hot_encode(self.labels, 10)      # 1-hot encoded labels
            self.N = self.x.shape[0]                           # number of datapoints

        @staticmethod
        def _dequantize(x, rng):
            """
            Adds noise to pixels to dequantize them.
            """
            return (x + rng.rand(*x.shape).astype(np.float32)) / 256.0

        @staticmethod
        def _logit_transform(x):
            """
            Transforms pixel values with logit to be unconstrained.
            """
            return logit(CIFAR10.alpha + (1 - 2*CIFAR10.alpha) * x)

        @staticmethod
        def _flip_augmentation(x):
            """
            Augments dataset x with horizontal flips.
            """
            D = int(x.shape[1] / 3)
            I = int(np.sqrt(D))
            r = x[:, :D].reshape([-1, I, I])[:, :, ::-1].reshape([-1, D])
            g = x[:, D:2*D].reshape([-1, I, I])[:, :, ::-1].reshape([-1, D])
            b = x[:, 2*D:].reshape([-1, I, I])[:, :, ::-1].reshape([-1, D])
            x_flip = np.hstack([r, g, b])
            return np.vstack([x, x_flip])

    def __init__(self, logit=False, flip=False, dequantize=True):

        rng = np.random.RandomState(42)

        # path = datasets.root + 'cifar10/'
#         path = "/hdd2/fatemeh/VRNN1/DensityEstimation/CIFAR/cifar-10-batches-py/"
        path = "/data3/fatemeh/server138/hdd2/VRNN1/DensityEstimation/CIFAR/cifar-10-batches-py/"

        # load train batches
        x = []
        l = []
        for i in range(1, 6):
            f = open(path + 'data_batch_' + str(i), 'rb')
            dict = pickle.load(f, encoding="bytes")
            x.append(dict[b'data'])
            l.append(dict[b'labels'])
            f.close()
        x = np.concatenate(x, axis=0)
        l = np.concatenate(l, axis=0)

        # use part of the train batches for validation
        split = int(0.9 * x.shape[0])
        self.trn = self.Data(x[:split], l[:split], logit, flip, dequantize, rng)
        self.val = self.Data(x[split:], l[split:], logit, False, dequantize, rng)

        # load test batch
        f = open(path + 'test_batch', 'rb')
        dict = pickle.load(f, encoding="bytes")
        x = dict[b'data']
        l = np.array(dict[b'labels'])
        f.close()
        self.tst = self.Data(x, l, logit, False, dequantize, rng)

        self.n_dims = self.trn.x.shape[1]
        self.n_labels = self.trn.y.shape[1]
        self.image_size = [int(np.sqrt(self.n_dims / 3))] * 2 + [3]