import torch
import torch.optim as optim
import argparse
import numpy as np
import sys
from mnist_data import MNIST
from cifar10_data import CIFAR10
from Networks import RMDN, gmm_sequence, RMDN_Uout, RMDN_Udirect, gmm_Uout, FRMDN
from scipy.special import logsumexp
import pickle
from functools import partial
import os

torch.manual_seed(0)

class EarlyStopping(object): # pylint: disable=R0902
    """
    Gives a criterion to stop training when a given metric is not
    improving anymore
    Args:
        mode (str): One of `min`, `max`. In `min` mode, training will
            be stopped when the quantity monitored has stopped
            decreasing; in `max` mode it will be stopped when the
            quantity monitored has stopped increasing. Default: 'min'.
        patience (int): Number of epochs with no improvement after
            which training is stopped. For example, if
            `patience = 2`, then we will ignore the first 2 epochs
            with no improvement, and will only stop learning after the
            3rd epoch if the loss still hasn't improved then.
            Default: 10.
        threshold (float): Threshold for measuring the new optimum,
            to only focus on significant changes. Default: 1e-4.
        threshold_mode (str): One of `rel`, `abs`. In `rel` mode,
            dynamic_threshold = best * ( 1 + threshold ) in 'max'
            mode or best * ( 1 - threshold ) in `min` mode.
            In `abs` mode, dynamic_threshold = best + threshold in
            `max` mode or best - threshold in `min` mode. Default: 'rel'.
    """

    # This function has been cloned from a repository.  (link)

    def __init__(self, mode='min', patience=100, threshold=0.01, threshold_mode='abs'):
        self.patience = patience
        self.mode = mode
        self.threshold = threshold
        self.threshold_mode = threshold_mode
        self.best = None
        self.num_bad_epochs = None
        self.mode_worse = None  # the worse value for the chosen mode
        self.is_better = None
        self.last_epoch = -1
        self._init_is_better(mode=mode, threshold=threshold,
                             threshold_mode=threshold_mode)
        self._reset()

    def _reset(self):
        """Resets num_bad_epochs counter and cooldown counter."""
        self.best = self.mode_worse
        self.num_bad_epochs = 0

    def step(self, metrics, epoch=None):
        """ Updates early stopping state """
        current = metrics
        if epoch is None:
            epoch = self.last_epoch = self.last_epoch + 1
        self.last_epoch = epoch

        if self.is_better(current, self.best):
            self.best = current
            self.num_bad_epochs = 0
        else:
            self.num_bad_epochs += 1

    @property
    def stop(self):
        """ Should we stop learning? """
        return self.num_bad_epochs > self.patience


    def _cmp(self, mode, threshold_mode, threshold, a, best): # pylint: disable=R0913, R0201
        if mode == 'min' and threshold_mode == 'rel':
            rel_epsilon = 1. - threshold
            return a < best * rel_epsilon

        elif mode == 'min' and threshold_mode == 'abs':
            return a < best - threshold

        elif mode == 'max' and threshold_mode == 'rel':
            rel_epsilon = threshold + 1.
            return a > best * rel_epsilon

        return a > best + threshold

    def _init_is_better(self, mode, threshold, threshold_mode):
        if mode not in {'min', 'max'}:
            raise ValueError('mode ' + mode + ' is unknown!')
        if threshold_mode not in {'rel', 'abs'}:
            raise ValueError('threshold mode ' + threshold_mode + ' is unknown!')

        if mode == 'min':
            self.mode_worse = float('inf')
        else:  # mode == 'max':
            self.mode_worse = (-float('inf'))

        self.is_better = partial(self._cmp, mode, threshold_mode, threshold)

    def state_dict(self):
        """ Returns early stopping state """
        return {key: value for key, value in self.__dict__.items() if key != 'is_better'}

    def load_state_dict(self, state_dict):
        """ Loads early stopping state """
        self.__dict__.update(state_dict)
        self._init_is_better(mode=self.mode, threshold=self.threshold,
                             threshold_mode=self.threshold_mode)

earlystopping = EarlyStopping() #('min', patience=10, threshold=0.1)

def validation(args, model, device, validation_loader, epoch):

    model.eval()

    L = []

    if (args.data_name == "mnist"):
        image_dim = 28
        sequence_length = 28
    else:
        image_dim = 32 * 3
        sequence_length = 32

    with torch.no_grad():
        max_iteration = len(validation_loader)
        for batch_idx, (data, target) in enumerate(validation_loader):

            if (batch_idx == max_iteration - 1):
                break

            data, target = data.float().to(device), target.float().to(device)
            data = data.view(-1, sequence_length, image_dim)
            add_noise = torch.from_numpy(np.random.normal(0, 0.01, (args.batch_size, args.class_number))).to(device)
            add_noise = add_noise.float()
            # print("########################")
            # print(target.shape)
            # print(add_noise.shape)
            target = target + add_noise
            target = target.view(-1, 1, args.class_number)
            target = target.repeat(1, sequence_length, 1)
            data_with_target = torch.cat((data, target), dim=2)
            data_with_target, data, target = data_with_target.to(device), data.to(device), target.to(device)

            if (args.flow == "yes"):
                zs, prior_logprob, log_det = model(data_with_target)
                loss = - prior_logprob - log_det
                loss = torch.mean(loss)
            else:
                if (args.cov == "simple"):
                    mu, sigma, logpi = model(data_with_target[:, :-1, :])
                    loss = gmm_sequence(data[:, 1:, :], mu, sigma, logpi)
                else:
                    mu, sigma, logpi, U = model(data_with_target[:, :-1, :])
                    loss = gmm_Uout(data[:, 1:, :], mu, sigma, logpi, U, args)

            L.append(loss.item())

    print("Validation = ", np.mean(L))

    return np.mean(L)

def train(args, model, device, train_loader, optimizer, epoch, fig_path):

    model.train()

    L = []

    if (args.data_name == "mnist"):
        image_dim = 28
        sequence_length = 28
    else:
        image_dim = 32 * 3
        sequence_length = 32

    max_iteration = len(train_loader)
    for batch_idx, (data, target) in enumerate(train_loader):

        if (batch_idx == max_iteration - 1):
            break

        data, target = data.float().to(device), target.float().to(device)
        data = data.view(-1, sequence_length, image_dim)
        add_noise = torch.from_numpy(np.random.normal(0, 0.01, (args.batch_size, args.class_number))).to(device)
        add_noise = add_noise.float()
        target = target + add_noise
        target = target.view(-1, 1, args.class_number)
        target = target.repeat(1, sequence_length, 1)
        data_with_target = torch.cat((data, target), dim=2)
        data_with_target, data, target = data_with_target.to(device), data.to(device), target.to(device)

        optimizer.zero_grad()
        if (args.flow == "yes"):
            zs, prior_logprob, log_det = model(data_with_target)
            loss = prior_logprob - log_det
            loss = torch.mean(loss)
            if ((epoch % 1 == 0) and (batch_idx == 0) and (args.cov == "simple")):
                model.sample(data_with_target, fig_path)
        else:
            if (args.cov == "simple"):
                mu, sigma, logpi = model(data_with_target[:, :-1, :])
                loss = gmm_sequence(data[:, 1:, :], mu, sigma, logpi)
                if ((epoch % 1 == 0) and (batch_idx == 0)):
                    model.sample(data_with_target, fig_path)
            else:
                mu, sigma, logpi, U = model(data_with_target[:, :-1, :])
                loss = gmm_Uout(data[:, 1:, :], mu, sigma, logpi, U, args)

        L.append(loss.item())
        loss.backward()
        optimizer.step()

        if batch_idx % 100 == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(train_loader.dataset),
                100. * batch_idx / len(train_loader), loss.item()))

    return np.mean(L)

def test(args, model, device, test_loader):

    model.eval()

    L = []

    if (args.data_name == "mnist"):
        image_dim = 28
        sequence_length = 28
    else:
        image_dim = 32 * 3
        sequence_length = 32

    cond_prob = []

    max_iteration = len(test_loader)
    with torch.no_grad():
        for batch_idx, (data, _) in enumerate(test_loader):

            if (batch_idx == max_iteration - 1):
                break

            data = data.float().to(device)
            data = data.view(-1, sequence_length, image_dim)
            add_noise = torch.from_numpy(np.random.normal(0, 0.01, (args.test_batch_size, args.class_number))).to(device)
            add_noise = add_noise.float()
            add_noise = add_noise.view(args.test_batch_size, 1, args.class_number)
            add_noise = add_noise.repeat(1, sequence_length, 1)
            target_onehot = torch.eye(args.class_number).to(device)

            conditional_p = np.zeros((args.class_number, args.test_batch_size))
            for class_number in range(args.class_number):
                target = target_onehot[class_number, :]
                target = target.view(1, 1, args.class_number)
                target = target.repeat(args.test_batch_size, sequence_length, 1)
                target = target + add_noise
                data_with_target = torch.cat((data, target), dim=2)
                data_with_target, data, target = data_with_target.to(device), data.to(device), target.to(device)

                if (args.flow == "yes"):
                    zs, prior_logprob, log_det = model(data_with_target)
                    loss = - (prior_logprob + log_det)
                elif (args.cov == "simple"):
                    mu, sigma, logpi = model(data_with_target[:, :-1, :])
                    loss = (-2.3026) + gmm_sequence(data[:, 1:, :], mu, sigma, logpi, reduce=False)
                else:
                    mu, sigma, logpi, U = model(data_with_target[:, :-1, :])
                    loss = (-2.3026) + gmm_Uout(data[:, 1:, :], mu, sigma, logpi, U, args, reduce=False)
                conditional_p[class_number, :] = loss.cpu().numpy()

            probability = logsumexp(conditional_p, axis=0)
            cond_prob += list(probability)


        print("conditional P = ", np.mean(cond_prob))

        L.append(np.mean(cond_prob))
        return np.mean(cond_prob)

def get_data(args):

    if (args.data_name == "mnist"):
        data = MNIST(logit=True, dequantize=True)
    elif (args.data_name == "cifar10"):
        data = CIFAR10(logit=True, flip=True, dequantize=True)
    else:
        print("\nError: wrong data")
        sys.exit(0)

    print("train = ", len(data.trn.x))
    print("test = ", len(data.tst.x))
    print("validation = ", len(data.val.x))

    train_data = []
    for i in range(len(data.trn.x)):
        train_data.append([data.trn.x[i, :], data.trn.y[i, :]])

    test_data = []
    for i in range(len(data.tst.x)):
        test_data.append([data.tst.x[i, :], data.tst.y[i, :]])

    validation_data = []
    for i in range(len(data.val.x)):
        validation_data.append([data.val.x[i, :], data.val.y[i, :]])

    train_loader = torch.utils.data.DataLoader(train_data, batch_size=args.batch_size, shuffle=True)
    test_loader = torch.utils.data.DataLoader(test_data, batch_size=args.test_batch_size, shuffle=False)
    validation_loader = torch.utils.data.DataLoader(validation_data, batch_size=args.test_batch_size, shuffle=False)

    return train_loader, test_loader, validation_loader

def main():

    parser = argparse.ArgumentParser("Density Estimation")
    parser.add_argument('--batch_size', type=int, default=128, help="batch size")
    parser.add_argument('--test_batch_size', type=int, default=128, help="test batch size")
    parser.add_argument('--epochs', type=int, default=5000, help='number of epochs')
    parser.add_argument('--lr', type=float, default=0.001, help='learning rate')
    parser.add_argument('--hidden_layer', type=int, default=1, help='lstm layer')
    parser.add_argument('--hidden_number', type=int, default=32, help='lstm hidden')
    parser.add_argument('--K', type=int, default=5, help='Gaussian number')
    parser.add_argument('--class_number', type=int, default=10, help='class number')
    parser.add_argument('--data_name', type=str, default="cifar10", help='mnist or cifar10')
    # parser.add_argument('--save_path', type=str, default="/hdd2/fatemeh/density/exp1/", help='save path')
    parser.add_argument('--save_path', type=str, default="/data1/fatemeh/hdd4/Thesis/cifar10/", help='save path')
    parser.add_argument('--flow', type=str, default="yes", help="with or without flow (yes or no)")
    parser.add_argument('--n_block', type=int, default=2, help="block number of flow, 2, 4, ...")
    parser.add_argument('--dprim', type=int, default=1, help='1, 2, 3')
    parser.add_argument('--cov', type=str, default="simple", help="simple, output, direct")
    args = parser.parse_args()


    name = ""
    name += "batch_size-{}".format(args.batch_size)
    name += "-test_batch_size-{}".format(args.test_batch_size)
    name += "-epochs-{}".format(args.epochs)
    name += "-lr-{}".format(args.lr)
    name += "-hidden_layer-{}".format(args.hidden_layer)
    name += "-hidden_number-{}".format(args.hidden_number)
    name += "-K-{}".format(args.K)
    name += "-class_number-{}".format(args.class_number)
    name += "-flow-{}".format(args.flow)
    name += "-nblobk-{}".format(args.n_block)
    name += "-dprim-{}".format(args.dprim)
    name += "-cov-{}".format(args.cov)
    print("name = ", name + "\n")


    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("device = ", device)


    train_loader, test_loader, validation_loader = get_data(args)


    if (args.flow == "no"):
        if (args.cov == "simple"):
            model = RMDN(args).to(device)
        elif (args.cov == "output"):
            model = RMDN_Uout(args).to(device)
        else:
            model = RMDN_Udirect(args).to(device)
    else:
        model = FRMDN(args).flow_based
        model = model.to(device)
    optimizer = optim.Adam(model.parameters(), lr=args.lr)


    print("\n")
    print(model)
    print("\n")


    if not (os.path.isdir(args.save_path + "/" + "figure")):
        os.mkdir(args.save_path + "/" + "figure")
    if not (os.path.isdir(args.save_path + "/" + "figure" + "/" + name)):
        os.mkdir(args.save_path + "/" + "figure" + "/" + name)
    # fig_path = args.save_path + args.data_name.upper() + "_generate" + "/" + "figure" + "/" + name + "-{}.png"
    fig_path = args.save_path + "/" + "figure" + "/" + name + "/" + "{}.png"

    LOSS = {}
    LOSS['train'] = []
    LOSS['test'] = []
    LOSS['validation'] = []
    LOSS['conditional_validation'] = []

    for epoch in range(1, args.epochs + 1):
        ltrain = train(args, model, device, train_loader, optimizer, epoch, fig_path.format(epoch))
        ltest = test(args, model, device, test_loader)
        LOSS['train'].append(ltrain)
        LOSS['test'].append(ltest)

        lvalidation = validation(args, model, device, validation_loader, epoch)
        LOSS['validation'].append(lvalidation)
        earlystopping.step(lvalidation)
        lconditional_validation = test(args, model, device, validation_loader)
        LOSS['conditional_validation'].append(lconditional_validation)

        file = open(args.save_path + "/" + name + ".pkl", "wb")
        pickle.dump(LOSS, file)
        file.close()
        torch.save(model, args.save_path + "/" + name + ".pt")

        if earlystopping.stop:
            print("End of Training because of early stopping at epoch {}".format(epoch))
            break

        print("\n")

main()
